@@ -5,6 +5,7 @@ property surface of a system.
 
 import itertools
 from collections import OrderedDict
+from collections.abc import Mapping
 import numpy as np
 from numpy import broadcast_to
 import pycalphad.variables as v
@@ -12,45 +13,43 @@ from pycalphad import ConditionError
 from pycalphad.codegen.callables import build_phase_records
 from pycalphad.core.cache import cacheit
 from pycalphad.core.light_dataset import LightDataset
+from pycalphad.model import Model
 from pycalphad.core.phase_rec import PhaseRecord
 from pycalphad.core.utils import endmember_matrix, extract_parameters, \
-    filter_phases, generate_dof, instantiate_models, point_sample, \
+    get_pure_elements, filter_phases, instantiate_models, point_sample, \
     unpack_components, unpack_condition, unpack_kwarg
 
 
 @cacheit
-def _sample_phase_constitution(phase_name, phase_constituents, sublattice_dof, comps,
-                               variables, sampler, fixed_grid, pdens):
+def _sample_phase_constitution(model, sampler, fixed_grid, pdens):
     """
     Sample the internal degrees of freedom of a phase.
 
     Parameters
     ----------
-    phase_name
-    phase_constituents
-    sublattice_dof
-    comps
-    variables
-    sampler
-    fixed_grid
-    pdens
+    model : Model
+        Instance of a pycalphad Model
+    sampler : Callable
+        Callable returning an ArrayLike of points
+    fixed_grid : bool
+        If True, sample pdens points between each pair of endmembers
+    pdens : int
+        Number of points to sample in each sampled dimension
 
     Returns
     -------
     ndarray of points
     """
     # Eliminate pure vacancy endmembers from the calculation
-    vacancy_indices = list()
-    for idx, sublattice in enumerate(phase_constituents):
-        active_in_subl = sorted(set(phase_constituents[idx]).intersection(comps))
-        is_vacancy = [spec.number_of_atoms == 0 for spec in active_in_subl]
-        subl_va_indices = list(idx for idx, x in enumerate(is_vacancy) if x == True)
+    vacancy_indices = []
+    for sublattice in model.constituents:
+        subl_va_indices = [idx for idx, spec in enumerate(sorted(set(sublattice))) if spec.number_of_atoms == 0]
         vacancy_indices.append(subl_va_indices)
-    if len(vacancy_indices) != len(phase_constituents):
+    if len(vacancy_indices) != len(model.constituents):
         vacancy_indices = None
+    sublattice_dof = [len(subl) for subl in model.constituents]
     # Add all endmembers to guarantee their presence
-    points = endmember_matrix(sublattice_dof,
-                              vacancy_indices=vacancy_indices)
+    points = endmember_matrix(sublattice_dof, vacancy_indices=vacancy_indices)
     if fixed_grid is True:
         # Sample along the edges of the endmembers
         # These constitution space edges are often the equilibrium points!
@@ -63,10 +62,8 @@ def _sample_phase_constitution(phase_name, phase_constituents, sublattice_dof, c
 
     # Sample composition space for more points
     if sum(sublattice_dof) > len(sublattice_dof):
-        points = np.concatenate((points,
-                                 sampler(sublattice_dof,
-                                         pdof=pdens)
-                                 ))
+        points = np.concatenate((points, sampler(sublattice_dof, pdof=pdens)))
+
     # Filter out nan's that may have slipped in if we sampled too high a vacancy concentration
     # Issues with this appear to be platform-dependent
     points = points[~np.isnan(points).any(axis=-1)]
@@ -230,7 +227,7 @@ def _compute_phase_values(components, statevar_dict,
     return LightDataset(data_arrays, coords=coordinate_dict)
 
 
-def calculate(dbf, comps, phases, mode=None, output='GM', fake_points=False, broadcast=True, parameters=None, to_xarray=True, **kwargs):
+def calculate(dbf, comps, phases, mode=None, output='GM', fake_points=False, broadcast=True, parameters=None, to_xarray=True, phase_records=None, **kwargs):
     """
     Sample the property surface of 'output' containing the specified
     components and phases. Model parameters are taken from 'dbf' and any
@@ -271,6 +268,11 @@ def calculate(dbf, comps, phases, mode=None, output='GM', fake_points=False, bro
         The density of points is determined by 'pdens'
     parameters : dict, optional
         Maps SymPy Symbol to numbers, for overriding the values of parameters in the Database.
+    phase_records : Optional[Mapping[str, PhaseRecord]]
+        Mapping of phase names to PhaseRecord objects. Must include all active phases.
+        The `model` argument must be a mapping of phase names to instances of Model
+        objects. Callers must take care that the PhaseRecord objects were created with
+        the same `output` as passed to `calculate`.
 
     Returns
     -------
@@ -287,6 +289,7 @@ def calculate(dbf, comps, phases, mode=None, output='GM', fake_points=False, bro
     callables = kwargs.pop('callables', {})
     sampler_dict = unpack_kwarg(kwargs.pop('sampler', None), default_arg=None)
     fixedgrid_dict = unpack_kwarg(kwargs.pop('grid_points', True), default_arg=True)
+    model = kwargs.pop('model', None)
     parameters = parameters or dict()
     if isinstance(parameters, dict):
         parameters = OrderedDict(sorted(parameters.items(), key=str))
@@ -298,6 +301,7 @@ def calculate(dbf, comps, phases, mode=None, output='GM', fake_points=False, bro
     if points_dict is None and broadcast is False:
         raise ValueError('The \'points\' keyword argument must be specified if broadcast=False is also given.')
     nonvacant_components = [x for x in sorted(comps) if x.number_of_atoms > 0]
+    nonvacant_elements = get_pure_elements(dbf, comps)
 
     all_phase_data = []
     largest_energy = 1e10
@@ -306,12 +310,10 @@ def calculate(dbf, comps, phases, mode=None, output='GM', fake_points=False, bro
     list_of_possible_phases = filter_phases(dbf, comps)
     if len(list_of_possible_phases) == 0:
         raise ConditionError('There are no phases in the Database that can be active with components {0}'.format(comps))
-    active_phases = {name: dbf.phases[name] for name in filter_phases(dbf, comps, phases)}
+    active_phases = filter_phases(dbf, comps, phases)
     if len(active_phases) == 0:
         raise ConditionError('None of the passed phases ({0}) are active. List of possible phases: {1}.'.format(phases, list_of_possible_phases))
 
-    models = instantiate_models(dbf, comps, list(active_phases.keys()), model=kwargs.pop('model', None), parameters=parameters)
-
     if isinstance(output, (list, tuple, set)):
         raise NotImplementedError('Only one property can be specified in calculate() at a time')
     output = output if output is not None else 'GM'
@@ -328,30 +330,49 @@ def calculate(dbf, comps, phases, mode=None, output='GM', fake_points=False, bro
     statevar_dict = dict((v.StateVariable(key), unpack_condition(value)) for key, value in kwargs.items() if key in statevar_strings)
     # Sort after default state variable check to fix gh-116
     statevar_dict = OrderedDict(sorted(statevar_dict.items(), key=lambda x: str(x[0])))
-    phase_records = build_phase_records(dbf, comps, active_phases, statevar_dict,
-                                   models=models, parameters=parameters,
-                                   output=output, callables=callables, build_gradients=False, build_hessians=False,
-                                   verbose=kwargs.pop('verbose', False))
     str_statevar_dict = OrderedDict((str(key), unpack_condition(value)) for (key, value) in statevar_dict.items())
+
+    # Build phase records if they weren't passed
+    if phase_records is None:
+        models = instantiate_models(dbf, comps, active_phases, model=model, parameters=parameters)
+        phase_records = build_phase_records(dbf, comps, active_phases, statevar_dict,
+                                            models=models, parameters=parameters,
+                                            output=output, callables=callables,
+                                            build_gradients=False, build_hessians=False,
+                                            verbose=kwargs.pop('verbose', False))
+    else:
+        # phase_records were provided, instantiated models must also be provided by the caller
+        models = model
+        if not isinstance(models, Mapping):
+            raise ValueError("A dictionary of instantiated models must be passed to `equilibrium` with the `model` argument if the `phase_records` argument is used.")
+        active_phases_without_models = [name for name in active_phases if not isinstance(models.get(name), Model)]
+        active_phases_without_phase_records = [name for name in active_phases if not isinstance(phase_records.get(name), PhaseRecord)]
+        if len(active_phases_without_phase_records) > 0:
+            raise ValueError(f"phase_records must contain a PhaseRecord instance for every active phase. Missing PhaseRecord objects for {sorted(active_phases_without_phase_records)}")
+        if len(active_phases_without_models) > 0:
+            raise ValueError(f"model must contain a Model instance for every active phase. Missing Model objects for {sorted(active_phases_without_models)}")
+
     maximum_internal_dof = max(len(models[phase_name].site_fractions) for phase_name in active_phases)
-    for phase_name, phase_obj in sorted(active_phases.items()):
+    for phase_name in sorted(active_phases):
         mod = models[phase_name]
         phase_record = phase_records[phase_name]
         points = points_dict[phase_name]
-        variables, sublattice_dof = generate_dof(phase_obj, mod.components)
         if points is None:
-            points = _sample_phase_constitution(phase_name, phase_obj.constituents, sublattice_dof, comps,
-                                                tuple(variables), sampler_dict[phase_name] or point_sample,
+            points = _sample_phase_constitution(mod, sampler_dict[phase_name] or point_sample,
                                                 fixedgrid_dict[phase_name], pdens_dict[phase_name])
         points = np.atleast_2d(points)
 
-        fp = fake_points and (phase_name == sorted(active_phases.keys())[0])
+        fp = fake_points and (phase_name == sorted(active_phases)[0])
         phase_ds = _compute_phase_values(nonvacant_components, str_statevar_dict,
                                          points, phase_record, output,
                                          maximum_internal_dof, broadcast=broadcast, parameters=parameters,
                                          largest_energy=float(largest_energy), fake_points=fp)
         all_phase_data.append(phase_ds)
 
+    fp_offset = len(nonvacant_elements) if fake_points else 0
+    running_total = [fp_offset] + list(np.cumsum([phase_ds['X'].shape[-2] for phase_ds in all_phase_data]))
+    islice_by_phase = {phase_name: slice(running_total[phase_idx], running_total[phase_idx+1], None)
+                       for phase_idx, phase_name in enumerate(sorted(active_phases))}
     # speedup for single-phase case (found by profiling)
     if len(all_phase_data) > 1:
         concatenated_coords = all_phase_data[0].coords
@@ -369,6 +390,7 @@ def calculate(dbf, comps, phases, mode=None, output='GM', fake_points=False, bro
         final_ds = LightDataset(data_vars=concatenated_data_vars, coords=concatenated_coords)
     else:
         final_ds = all_phase_data[0]
+    final_ds.attrs['phase_indices'] = islice_by_phase
     if to_xarray:
         return final_ds.get_dataset()
     else:
